__all__ = ["ManagedPolicyOnUserRule"]

from typing import Dict, Optional

from pycfmodel.model.cf_model import CFModel
from pycfmodel.model.resources.iam_managed_policy import IAMManagedPolicy

from cfripper.model.enums import RuleGranularity
from cfripper.model.result import Result
from cfripper.rules.base_rules import Rule


class ManagedPolicyOnUserRule(Rule):
    """
    Checks if any IAM managed policy is applied to a group and not a user.

    Risk:
        Instead of defining permissions for individual IAM users, it's usually more convenient and secure
        to create [IAM groups](https://docs.aws.amazon.com/IAM/latest/UserGuide/id_groups.html) that relate
        to different functions. IAM users can be assigned to these groups.
        All the users in an IAM group inherit the permissions assigned to the group. That way, you can make
        changes for everyone in a group in just one place. As people move around in your company, you can
        simply change what IAM group their IAM user belongs to, without risking a user having too much
        privilege.

    Fix:
        Use IAM Groups as opposed to users in IAM Managed Policies.

    Code for fix:
        This is an example which will be flagged by CFRipper:

        ```json
        "BadPolicy": {
          "Type": "AWS::IAM::ManagedPolicy",
          "Properties": {
            "Description": "Policy for something.",
            "Path": "/",
            "PolicyDocument": {
              "Version": "2012-10-17",
              "Statement": [...]
            },
            "Users": [{"Ref": "TestUser"}]
          }
        }
        ```

        This is an example of a more acceptable CloudFormation policy:

        ```json
        "GoodPolicy": {
          "Type": "AWS::IAM::ManagedPolicy",
          "Properties": {
            "Description": "Policy for something.",
            "Path": "/",
            "PolicyDocument": {
              "Version": "2012-10-17",
              "Statement": [...]
            },
            "Groups": ["user_group"]
          }
        }
        ```

    Filters context:
        | Parameter               | Type                             | Description                                                    |
        |:-----------------------:|:--------------------------------:|:--------------------------------------------------------------:|
        |`config`                 | str                              | `config` variable available inside the rule                    |
        |`extras`                 | str                              | `extras` variable available inside the rule                    |
        |`logical_id`             | str                              | ID used in Cloudformation to refer the resource being analysed |
        |`resource`               | `IAMManagedPolicy`               | Resource that is being addressed                               |
    """

    REASON = "IAM managed policy {} should not apply directly to users. Should be on group"
    GRANULARITY = RuleGranularity.RESOURCE

    def invoke(self, cfmodel: CFModel, extras: Optional[Dict] = None) -> Result:
        result = Result()
        for logical_id, resource in cfmodel.Resources.items():
            if isinstance(resource, IAMManagedPolicy) and resource.Properties.Users:
                self.add_failure_to_result(
                    result,
                    self.REASON.format(logical_id),
                    resource_ids={logical_id},
                    context={"config": self._config, "extras": extras, "logical_id": logical_id, "resource": resource},
                )
        return result



import logging
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Set, Tuple, Type

from pycfmodel.model.cf_model import CFModel
from pycfmodel.model.resources.resource import Resource

from cfripper.config.config import Config
from cfripper.config.filter import Filter
from cfripper.config.rule_config import RuleConfig
from cfripper.model.enums import RuleGranularity, RuleMode, RuleRisk
from cfripper.model.result import Result

logger = logging.getLogger(__file__)


class Rule(ABC):
    RULE_MODE = RuleMode.BLOCKING
    RISK_VALUE = RuleRisk.MEDIUM
    GRANULARITY = RuleGranularity.STACK

    def __init__(self, config: Optional[Config]):
        self._config = config if config else Config()

    @property
    def rule_config(self) -> RuleConfig:
        return self._config.get_rule_config(self.__class__.__name__)

    @property
    def rule_filters(self) -> List[Filter]:
        return self._config.get_rule_filters(self.__class__.__name__)

    @property
    def rule_mode(self) -> RuleMode:
        return self.rule_config.rule_mode or self.RULE_MODE

    @property
    def risk_value(self) -> RuleRisk:
        return self.rule_config.risk_value or self.RISK_VALUE

    @abstractmethod
    def invoke(self, cfmodel: CFModel, extras: Optional[Dict] = None) -> Result:
        pass

    def add_failure_to_result(
        self,
        result: Result,
        reason: str,
        granularity: Optional[RuleGranularity] = None,
        resource_ids: Optional[Set] = None,
        actions: Optional[Set] = None,
        risk_value: Optional[RuleRisk] = None,
        rule_mode: Optional[RuleMode] = None,
        context: Optional[Dict] = None,
    ):
        rule_mode = rule_mode or self.rule_mode
        risk_value = risk_value or self.risk_value
        granularity = granularity or self.GRANULARITY

        for fltr in self.rule_filters:
            try:
                if fltr(**context):
                    risk_value = fltr.risk_value or risk_value
                    rule_mode = fltr.rule_mode or rule_mode
            except Exception:
                logger.exception(f"Exception raised while evaluating filter for `{fltr.reason}`", extra=context)

        if rule_mode != RuleMode.ALLOWED:
            result.add_failure(
                rule=type(self).__name__,
                reason=reason,
                rule_mode=rule_mode,
                risk_value=risk_value,
                resource_ids=resource_ids,
                actions=actions,
                granularity=granularity,
            )


class ResourceSpecificRule(Rule):
    """
    Base class for rules that only apply to a subset of resource types.

    RESOURCE_TYPES: Resources to invoke the rule for.
    EXCLUDED_RESOURCE_TYPES: Resources to explicitly not run the rule for.

    Both fields are included to allow for more granular rule definitions. For example,
    you may want to allow all resources except S3BucketPolicies, in which case you
    would define these variables as:

    EXCLUDED_RESOURCE_TYPES = (S3BucketPolicy,)
    RESOURCE_TYPES = (Resource,)

    Where the `S3BucketPolicy` Resource inherits from the base `Resource` class.
    """

    EXCLUDED_RESOURCE_TYPES: Tuple[Type] = tuple()
    RESOURCE_TYPES: Tuple[Type] = tuple()

    def invoke(self, cfmodel: CFModel, extras: Optional[Dict] = None) -> Result:
        result = Result()
        for logical_id, resource in cfmodel.Resources.items():
            if isinstance(resource, self.RESOURCE_TYPES) and not isinstance(resource, self.EXCLUDED_RESOURCE_TYPES):
                result += self.resource_invoke(resource=resource, logical_id=logical_id, extras=extras)
        return result

    @abstractmethod
    def resource_invoke(self, resource: Resource, logical_id: str, extras: Optional[Dict] = None) -> Result:
        pass


class PrincipalCheckingRule(Rule, ABC):
    """
    Abstract class for rules that check principals.

    `valid_principals` is a set of the following Account IDs and Canonical IDs:
      - `aws_principals` set in the user defined config (default = None)
      - ELB Log Account IDs from AWS
      - ElastiCache Backup Canonical IDs
      - (if defined) The AWS Account in the config which CFRipper is executing with

    When using `valid_principals`, make sure the scope of accounts allowed is not too large.
    It might be the case that the account the stack is being deployed in is in this set.
    This could raise false negatives in rules. If a rule should only be exempt for AWS Service
    IDs, such as ELB and ElastiCache, consider using `_get_allowed_from_config()` directly.
    """

    _valid_principals = None

    def _get_allowed_from_config(self, services: List[str] = None) -> Set[str]:
        if services is None:
            services = self._config.aws_service_accounts.keys()

        unique_list = set()
        for service in services:
            unique_list |= set(self._config.aws_service_accounts[service])
        return unique_list

    @property
    def valid_principals(self) -> Set[str]:
        if self._valid_principals is None:
            self._valid_principals = {
                *self._config.aws_principals,
                *self._get_allowed_from_config(),
            }
            if self._config.aws_account_id:
                self._valid_principals.add(self._config.aws_account_id)
        return self._valid_principals


class BaseDangerousPolicyActions(ResourceSpecificRule, ABC):
    """
    Base class for dangerous actions. Admits a DANGEROUS_ACTIONS class variable with a list of dangerous actions
    """

    DEFAULT_FILTERS_CONTEXT = """\
    Filters context:
        | Parameter    | Type             | Description                                                     |
        |:------------:|:----------------:|:---------------------------------------------------------------:|
        |`config`      | str              | `config` variable available inside the rule                     |
        |`extras`      | str              | `extras` variable available inside the rule                     |
        |`logical_id`  | str              | ID used in Cloudformation to refer the resource being analysed  |
        |`policy_name` | `Optional[str]`  | If available, the policy name                                   |
        |`action`      | `List[str]`      | List of dangerous actions contained within the policy           |
    """

    REASON = "Resource {} should not include the following dangerous actions: {}"
    RISK_VALUE = RuleRisk.HIGH
    GRANULARITY = RuleGranularity.ACTION

    @property
    @classmethod
    @abstractmethod
    def DANGEROUS_ACTIONS(cls) -> List[str]:
        # This is designed to be overwritten as a class variable
        raise NotImplementedError

    def resource_invoke(self, resource: Resource, logical_id: str, extras: Optional[Dict] = None) -> Result:
        result = Result()
        for policy in resource.policy_documents:
            actions = policy.policy_document.get_allowed_actions()
            dangerous_actions = set(actions) & set(self.DANGEROUS_ACTIONS)
            if dangerous_actions:
                self.add_failure_to_result(
                    result,
                    self.REASON.format(logical_id, sorted(dangerous_actions)),
                    resource_ids={logical_id},
                    actions=dangerous_actions,
                    context={
                        "config": self._config,
                        "extras": extras,
                        "logical_id": logical_id,
                        "policy_name": policy.name,
                        "actions": sorted(dangerous_actions),
                    },
                )
        return result



import json
import logging
import re
from contextlib import suppress
from functools import lru_cache
from typing import Optional
from urllib.parse import unquote

import boto3
import yaml
from cfn_flip import to_json
from pycfmodel.model.resources.properties.policy import Policy

from cfripper.config.regex import REGEX_ARN, REGEX_IAM_ARN, REGEX_STS_ARN

logger = logging.getLogger(__file__)


class InvalidURLException(Exception):
    pass


def extract_bucket_name_and_path_from_url(url):
    # Remove query string
    url = unquote(url).split("?")[0]

    bucket_name = None
    path = None

    # https://bucket.s3.amazonaws.com/path1/path2
    match = re.search(r"^https://([^.]+)\.s3\.amazonaws\.com(.*?)$", url)
    if match:
        bucket_name, path = match.group(1), match.group(2)[1:]  # Trim start /

    # https://bucket.s3.aws-region.amazonaws.com/path1/path2
    match = re.search(r"^https://([^.]+)\.s3\.[^\.]+\.amazonaws\.com(.*?)$", url)
    if match:
        bucket_name, path = match.group(1), match.group(2)[1:]  # Trim start /

    # https://bucket.s3-aws-region.amazonaws.com/path1/path2
    match = re.search(r"^https://([^.]+)\.s3-[^\.]+\.amazonaws\.com(.*?)$", url)
    if match:
        bucket_name, path = match.group(1), match.group(2)[1:]  # Trim start /

    # https://s3.amazonaws.com/bucket/path1/path2
    match = re.search(r"^https://s3\.amazonaws\.com/([^\/]+)(.*?)$", url)
    if match:
        bucket_name, path = match.group(1), match.group(2)[1:]  # Trim start /

    # https://s3.aws-region.amazonaws.com/bucket/path1/path2
    match = re.search(r"^https://s3\.[^.]+\.amazonaws\.com/([^\/]+)(.*?)$", url)
    if match:
        bucket_name, path = match.group(1), match.group(2)[1:]  # Trim start /

    # https://s3-aws-region.amazonaws.com/bucket/path1/path2
    match = re.search(r"^https://s3-[^.]+\.amazonaws\.com/([^\/]+)(.*?)$", url)
    if match:
        bucket_name, path = match.group(1), match.group(2)[1:]  # Trim start /

    if bucket_name is None and path is None:
        raise InvalidURLException(f"Couldn't extract bucket name and path from url: {url}")

    logger.info(f"extract_bucket_name_and_path_from_url. returning for {bucket_name} and {path} for {url}")

    return bucket_name, path


def convert_json_or_yaml_to_dict(file_content, stack_id: Optional[str] = None):
    with suppress(ValueError):
        return json.loads(file_content)

    try:
        # Convert file_content (assuming that is YAML) to JSON if possible
        file_content = to_json(file_content)
        return json.loads(file_content)
    except yaml.YAMLError:
        logger.exception(f"Could not convert YAML to JSON template for {stack_id}")
    except ValueError:
        logger.exception(f"Could not parse JSON template for {stack_id}")

    return None


@lru_cache(maxsize=None)
def get_managed_policy(managed_policy_arn):
    iam_client = boto3.client("iam")
    managed_policy = iam_client.get_policy(PolicyArn=managed_policy_arn)
    version_id = managed_policy.get("Policy", {}).get("DefaultVersionId")
    if version_id:
        policy_version = iam_client.get_policy_version(PolicyArn=managed_policy_arn, VersionId=version_id)
        return Policy(
            **{
                "PolicyDocument": policy_version["PolicyVersion"]["Document"],
                "PolicyName": f"AutoTransformedManagedPolicy{version_id}",
            }
        )
    return None


def get_aws_service_from_arn(arn: str) -> Optional[str]:
    match = REGEX_ARN.match(arn)
    if match:
        return match.group(1)


def get_account_id_from_arn(arn: str) -> Optional[str]:
    match = REGEX_ARN.match(arn)
    if match:
        return match.group(3)


def get_account_id_from_iam_arn(arn: str) -> Optional[str]:
    match = REGEX_IAM_ARN.match(arn)
    if match:
        return match.group(1)


def get_account_id_from_sts_arn(arn: str) -> Optional[str]:
    match = REGEX_STS_ARN.match(arn)
    if match:
        return match.group(1)


def get_account_id_from_principal(principal: str) -> Optional[str]:
    if principal.isnumeric():
        return principal

    aws_service = get_aws_service_from_arn(principal)
    if aws_service not in ["iam", "sts"]:
        return None

    if aws_service == "iam":
        return get_account_id_from_iam_arn(principal)
    elif aws_service == "sts":
        return get_account_id_from_sts_arn(principal)



__all__ = ["GenericWildcardPrincipalRule", "PartialWildcardPrincipalRule", "FullWildcardPrincipalRule"]
import logging
import re
from typing import Dict, Optional

from pycfmodel.model.cf_model import CFModel
from pycfmodel.model.resources.iam_managed_policy import IAMManagedPolicy
from pycfmodel.model.resources.iam_policy import IAMPolicy
from pycfmodel.model.resources.iam_role import IAMRole
from pycfmodel.model.resources.iam_user import IAMUser
from pycfmodel.model.resources.properties.policy_document import PolicyDocument
from pycfmodel.model.resources.s3_bucket_policy import S3BucketPolicy
from pycfmodel.model.resources.sns_topic_policy import SNSTopicPolicy
from pycfmodel.model.resources.sqs_queue_policy import SQSQueuePolicy

from cfripper.config.regex import REGEX_FULL_WILDCARD_PRINCIPAL, REGEX_PARTIAL_WILDCARD_PRINCIPAL
from cfripper.model.enums import RuleGranularity, RuleRisk
from cfripper.model.result import Result
from cfripper.rules.base_rules import PrincipalCheckingRule

logger = logging.getLogger(__file__)


class GenericWildcardPrincipalRule(PrincipalCheckingRule):
    REASON_WILDCARD_PRINCIPAL = "{} should not allow wildcards in principals (principal: '{}')"
    GRANULARITY = RuleGranularity.RESOURCE

    AWS_ACCOUNT_ID_PATTERN = re.compile(r"^(\d{12})$")
    IAM_PATTERN = re.compile(r"arn:aws:iam::(\d*|\*):.*")
    FULL_REGEX = REGEX_FULL_WILDCARD_PRINCIPAL

    def invoke(self, cfmodel: CFModel, extras: Optional[Dict] = None) -> Result:
        result = Result()
        for logical_id, resource in cfmodel.Resources.items():
            if isinstance(resource, (IAMManagedPolicy, IAMPolicy, S3BucketPolicy, SNSTopicPolicy, SQSQueuePolicy)):
                self.check_for_wildcards(result, logical_id, resource.Properties.PolicyDocument, extras)
            elif isinstance(resource, (IAMRole, IAMUser)):
                if isinstance(resource, IAMRole):
                    self.check_for_wildcards(result, logical_id, resource.Properties.AssumeRolePolicyDocument, extras)
                if resource.Properties and resource.Properties.Policies:
                    for policy in resource.Properties.Policies:
                        self.check_for_wildcards(result, logical_id, policy.PolicyDocument, extras)
        return result

    def check_for_wildcards(
        self, result: Result, logical_id: str, resource: PolicyDocument, extras: Optional[Dict] = None
    ):
        for statement in resource._statement_as_list():
            if statement.Effect == "Allow" and statement.principals_with(self.FULL_REGEX):
                for principal in statement.get_principal_list():
                    account_id_match = self.IAM_PATTERN.match(principal) or self.AWS_ACCOUNT_ID_PATTERN.match(principal)
                    account_id = account_id_match.group(1) if account_id_match else None

                    # Check if account ID is allowed. `self._get_allowed_from_config()` used here
                    # to reduce number of false negatives and only allow exemptions for accounts
                    # which belong to AWS Services (such as ELB and ElastiCache).
                    if account_id in self._get_allowed_from_config():
                        continue

                    if statement.Condition and statement.Condition.dict():
                        logger.warning(
                            f"Not adding {type(self).__name__} failure in {logical_id} because there are conditions: "
                            f"{statement.Condition}"
                        )
                    else:
                        self.add_failure_to_result(
                            result,
                            self.REASON_WILDCARD_PRINCIPAL.format(logical_id, principal),
                            resource_ids={logical_id},
                            context={
                                "config": self._config,
                                "extras": extras,
                                "logical_id": logical_id,
                                "resource": resource,
                                "statement": statement,
                                "principal": principal,
                                "account_id": account_id,
                            },
                        )


class PartialWildcardPrincipalRule(GenericWildcardPrincipalRule):
    """
    Checks for any wildcard or account-wide principals defined in any statements. This rule will flag
    as non-compliant any principals where `root` or `*` are included at the end of the value, for
    example, `arn:aws:iam:12345:12345*`.

    Risk:
        It might allow other AWS identities or the root access of the account to escalate privileges.

    Fix:
        Where possible, restrict the access to only the required resources.
        For example, instead of `Principal: "*"`, include a list of the roles that need access.

    Filters context:
        | Parameter   | Type               | Description                                                    |
        |:-----------:|:------------------:|:--------------------------------------------------------------:|
        |`config`     | str                | `config` variable available inside the rule                    |
        |`extras`     | str                | `extras` variable available inside the rule                    |
        |`logical_id` | str                | ID used in Cloudformation to refer the resource being analysed |
        |`resource`   | `S3BucketPolicy`   | Resource that is being addressed                               |
        |`statement`  | `Statement`        | Statement being checked found in the Resource                  |
        |`principal`  | str                | AWS Principal being checked found in the statement             |
        |`account_id` | str                | Account ID found in the principal                              |
    """

    REASON_WILDCARD_PRINCIPAL = (
        "{} should not allow wildcard in principals or account-wide principals (principal: '{}')"
    )
    RISK_VALUE = RuleRisk.MEDIUM
    FULL_REGEX = REGEX_PARTIAL_WILDCARD_PRINCIPAL


class FullWildcardPrincipalRule(GenericWildcardPrincipalRule):
    """
    Checks for any wildcard principals defined in any statements.

    Risk:
        It might allow other AWS identities to escalate privileges.

    Fix:
        Where possible, restrict the access to only the required resources.
        For example, instead of `Principal: "*"`, include a list of the roles that need access.

    Filters context:
        | Parameter   | Type               | Description                                                    |
        |:-----------:|:------------------:|:--------------------------------------------------------------:|
        |`config`     | str                | `config` variable available inside the rule                    |
        |`extras`     | str                | `extras` variable available inside the rule                    |
        |`logical_id` | str                | ID used in Cloudformation to refer the resource being analysed |
        |`resource`   | `S3BucketPolicy`   | Resource that is being addressed                               |
        |`statement`  | `Statement`        | Statement being checked found in the Resource                  |
        |`principal`  | str                | AWS Principal being checked found in the statement             |
        |`account_id` | str                | Account ID found in the principal                              |
    """

    RISK_VALUE = RuleRisk.HIGH
